# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class GapConditions:
    def gp(self, node1="", node2="", lab="", stif="", gap="", damp="", **kwargs):
        """Defines a gap condition for transient analyses.

        APDL Command: GP

        Parameters
        ----------
        node1
            Node I of gap.  If NODE1 = P, graphical picking is enabled and all
            remaining command fields are ignored (valid only in the GUI).

        node2
            Node J of gap (must be different from NODE1).  Non-grounded gap
            nodes must be defined as master degrees of freedom or be
            unconstrained, active DOF in a full analysis type.  Grounded gap
            nodes (those not defined as MDOF) need not appear elsewhere in the
            model.

        lab
            Direction of gap action in the nodal coordinate system (implied
            from the following force labels): FX, FY, FZ, MX, MY, MZ.

        stif
            Stiffness (Force/Length) of closed gap (may be positive or
            negative).

        gap
            Initial size of gap.  A zero (or positive) value assumes an
            initially open gap.  A negative value defines an interference
            condition.  For a rotational gap, GAP should be in radians.

        damp
            Damping coefficient (``Force*Time/Length``) of closed gap using pseudo
            velocity (Newmark finite difference expansion scheme).

        Notes
        -----
        Defines a gap condition for the mode superposition transient analysis
        (ANTYPE,TRANS with TRNOPT,MSUP). If used in SOLUTION, this command is
        valid only within the first load step. Gap conditions specified in
        subsequent load steps are ignored.

        Repeat GP command for additional gap conditions. Gaps are numbered
        sequentially as input.

        Note:: : Gaps may be renumbered by the program during the solution (see
        output listing)

        The mode-superposition transient analysis does not allow gap action
        with the standard gap elements. However, you can define gap conditions
        which are similar to gap elements; gap conditions can be specified
        between surfaces that are expected to contact (impact) each other
        during the transient. The gap condition simulates the basic gap action
        of the COMBIN40 element.

        The gap condition is treated as an explicit force (equal to the
        interference times contact stiffness) and affects only the load vector
        calculation and not the stiffness matrix. The interference is
        calculated from the displacement extrapolated from the previous time
        points.

        Gap conditions can only be defined between two master degree of freedom
        (DOF) nodes or between master DOF nodes and ground, as shown in the
        following figure.

        Master degrees of freedom are the unconstrained and active degrees of
        freedom. Gap nodes not defined as active degrees of freedom or attached
        to an element are assumed to be grounded. Grounded gap nodes do not
        need a spatial location, nor do they need to be located on an element.

        Gap conditions may be defined in parallel (across the same nodes), with
        varying gap and stiffness values, to simulate a nonlinear (piecewise)
        force-deflection curve.

        The gap direction is determined from the force label input on the GP
        command; i.e., FX defines a translational gap acting in the UX nodal
        degree of freedom direction, and MZ defines a rotational gap acting in
        the nodal ROTZ degree of freedom direction. The actual degree of
        freedom directions available for a particular node depends upon the
        degrees of freedom associated with the element types [ET] at that node.

        If the coordinate systems of the nodes connecting the gap are rotated
        relative to each other, the same degree of freedom may be in different
        directions. The gap, however, assumes only a one-dimensional action.
        Nodes I and J may be anywhere in space (preferably coincident).  No
        moment effects are included due to noncoincident nodes. That is, if the
        nodes are offset from the line of action, moment equilibrium may not be
        satisfied.

        The contact stiffness value represents the stiffness of the closed gap.
        Stiffness values are related to the integration time step size and
        should be physically reasonable. High stiffness will require a small
        integration time step; otherwise, due to the displacement
        extrapolation, the solution may go unstable. Negative stiffness values
        may be used with gaps in parallel to produce a decreasing force-
        deflection curve.

        The order of specifying the gap nodes is important; i.e., a gap
        condition connecting two nodes will act differently depending upon
        which node is specified first on the GP command.  For example, for Node
        1 at X = 0.0, Node 2 at X = 0.1, and the gap defined from Node 1 to 2,
        a displacement of Node 1 greater than Node 2 will cause the gap to
        close.  For the gap defined from Node 2 to 1, a displacement of Node 2
        greater than Node 1 will cause the gap to close (like a hook action).
        In general, the gap closes whenever the separation (defined as UJ - UI
        + GAP) is negative.  UJ is the displacement of node J, UI is the
        displacement of node I, and GAP is the input gap value.  The gap force
        output appears in the printout only for the time steps for which the
        gap is closed.  A negative spring force is always associated with a
        closed gap (even with the hook option).

        Some guidelines to define gap conditions are presented below:

        Use enough gap conditions to obtain a smooth contact stress
        distribution between the contacting surfaces.

        Define a reasonable gap stiffness. If the stiffness is too low, the
        contacting surfaces may overlap too much. If the stiffness is too high,
        a very small time step will be required during impact. A general
        recommendation is to specify a gap stiffness that is one or two orders
        of magnitude higher than the adjacent element stiffness. You can
        estimate the adjacent element stiffness using AE/L, where A is the
        contributing area around the gap condition, E is the elastic modulus of
        the softer material at the interface, and L is the depth of the first
        layer of elements at the interface.

        A mode-superposition transient using the nonlinear gap damping provided
        through the DAMP field runs faster than a full transient analysis using
        a gap element (COMBIN40).

        Use the GPLIST command to list gap conditions and the GPDELE command to
        delete gap conditions.

        This command is also valid in PREP7.
        """
        command = f"GP,{node1},{node2},{lab},{stif},{gap},{damp}"
        return self.run(command, **kwargs)

    def gpdele(self, gap1="", gap2="", ginc="", **kwargs):
        """Deletes gap conditions.

        APDL Command: GPDELE

        Parameters
        ----------
        gap1, gap2, ginc
            Delete gap conditions from GAP1 to GAP2 (defaults to GAP1) in steps
            of GINC  (defaults to 1).

        Notes
        -----
        Deletes gap conditions defined with the GP command.  Gap conditions
        following those deleted are automatically compressed and renumbered.
        If used in SOLUTION, this command is valid only within the first load
        step.

        This command is also valid in PREP7.
        """
        command = f"GPDELE,{gap1},{gap2},{ginc}"
        return self.run(command, **kwargs)

    def gplist(self, gap1="", gap2="", ginc="", **kwargs):
        """Lists the gap conditions.

        APDL Command: GPLIST

        Parameters
        ----------
        gap1, gap2, ginc
            List gap conditions from GAP1 to GAP2 (GAP2 defaults to GAP1) in
            steps of GINC  (defaults to 1).  If GAP1 = ALL (default), GAP2 and
            GINC  are ignored and all gap conditions are listed.

        Notes
        -----
        This command is valid in any processor.
        """
        command = f"GPLIST,{gap1},{gap2},{ginc}"
        return self.run(command, **kwargs)



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import re

from ansys.mapdl.core._commands import parse


class Nodes:
    def center(self, node="", node1="", node2="", node3="", radius="", **kwargs):
        """Defines a node at the center of curvature of 2 or 3 nodes.

        APDL Command: CENTER

        Parameters
        ----------
        node
            Number to be assigned to the node generated at the center of
            curvature.

        node1, node2, node3
            Three nodes used to calculated the center of curvature, as
            described under RADIUS.

        radius
            Used to control the interpretation of NODE1, NODE2 and NODE3:

            0 - NODE1, NODE2 and NODE3 lie on a circular arc.  The program will calculate the
                center of curvature (and radius) (default).

            ≠0 - NODE1 and NODE2 are the endpoints of an arc, and RADIUS is the radius of
                 curvature.  The program will locate the center of curvature on
                 the NODE3 side of the NODE1-NODE2 line if RADIUS > 0, and
                 opposite to NODE3 if RADIUS < 0.
        """
        command = f"CENTER,{node},{node1},{node2},{node3},{radius}"
        return self.run(command, **kwargs)

    def fill(
        self,
        node1="",
        node2="",
        nfill="",
        nstrt="",
        ninc="",
        itime="",
        inc="",
        space="",
        **kwargs,
    ):
        """Generates a line of nodes between two existing nodes.

        APDL Command: FILL

        Parameters
        ----------
        node1, node2
            Beginning and ending nodes for fill-in.  NODE1 defaults to next to
            last node specified, NODE2 defaults to last node specified.  If
            NODE1 = P, graphical picking is enabled and all remaining command
            fields are ignored (valid only in the GUI).

        nfill
            Fill NFILL nodes between NODE1 and NODE2 (defaults to
            ``|NODE2-NODE1|-1``).  NFILL must be positive.

        nstrt
            Node number assigned to first filled-in node (defaults to ``NODE1 +
            NINC``).

        ninc
            Add this increment to each of the remaining filled-in node numbers
            (may be positive or negative).  Defaults to the integer result of
            ``(NODE2-NODE1)/(NFILL + 1)``, i.e., linear interpolation.  If the
            default evaluates to zero, or if zero is input, NINC is set to 1.

        itime, inc
            Do fill-in operation a total of ITIMEs, incrementing NODE1, NODE2
            and NSTRT by INC each time after the first.  ITIME and INC both
            default to 1.

        space
            Spacing ratio.  Ratio of last division size to first division size.
            If > 1.0, divisions increase.  If < 1.0, divisions decrease.  Ratio
            defaults to 1.0 (uniform spacing).

        Notes
        -----
        Generates a line of nodes (in the active coordinate system) between two
        existing nodes.  The two nodes may have been defined in any coordinate
        system.  Nodal locations and rotation angles are determined by
        interpolation.  Any number of nodes may be filled-in and any node
        number sequence may be assigned.  See the CSCIR command when filling
        across the 180° singularity line in a non-Cartesian system.
        """
        command = f"FILL,{node1},{node2},{nfill},{nstrt},{ninc},{itime},{inc},{space}"
        return self.run(command, **kwargs)

    def move(
        self,
        node="",
        kc1="",
        x1="",
        y1="",
        z1="",
        kc2="",
        x2="",
        y2="",
        z2="",
        **kwargs,
    ):
        """Calculates and moves a node to an intersection.

        APDL Command: MOVE

        Parameters
        ----------
        node
            Move this node.  If NODE = P, graphical picking is enabled and all
            remaining command fields are ignored (valid only in the GUI).  A
            component name may also be substituted for NODE.

        kc1
            First coordinate system number.  Defaults to 0 (global Cartesian).

        x1, y1, z1
            Input one or two values defining the location of the node in this
            coordinate system.  Input "U" for unknown value(s) to be calculated
            and input "E" to use an existing coordinate value.  Fields are R1,
            θ1, Z1 for cylindrical, or R1, θ1, Φ1 for spherical or toroidal.

        kc2
            Second coordinate system number.

        x2, y2, z2
            Input two or one value(s) defining the location of the node in this
            coordinate system.  Input "U" for unknown value(s) to be calculated
            and input "E" to use an existing coordinate value.  Fields are R2,
            θ2, Z2 for cylindrical, or R2,  θ2, Φ2 for spherical or toroidal.

        Notes
        -----
        Calculates and moves a node to an intersection location.  The node may
        have been previously defined (at an approximate location) or left
        undefined (in which case it is internally defined at the SOURCE
        location).  The actual location is calculated from the intersection of
        three surfaces (implied from three coordinate constants in two
        different coordinate systems).  The three (of six) constants easiest to
        define should be used.  The program will calculate the remaining three
        coordinate constants.  All arguments, except KC1, must be input.  Use
        the repeat command [``*REPEAT``] after the MOVE command to define a line of
        intersection by repeating the move operation on all nodes of the line.

        Surfaces of constant value are implied by some commands by specifying a
        single coordinate value.  Implied surfaces are used with various
        commands [MOVE, KMOVE, NSEL, etc.].  Three surfaces are available with
        each of the four coordinate system types.  Values or X, Y, or Z may be
        constant for the Cartesian coordinate system; values of R,: θ, or Z for
        the cylindrical system; and values of R, θ,: Φ for the spherical and
        toroidal systems.  For example, an X value of 3 represents the Y-Z
        plane (or surface) at X=3.  In addition, the parameters for the
        cylindrical and spherical coordinate systems may be adjusted [CS,
        LOCAL] to form elliptical surfaces.  For surfaces in elliptical
        coordinate systems, a surface of "constant" radius is defined by the
        radius value at the X-axis.  Surfaces of constant value may be located
        in local coordinate systems [LOCAL, CLOCAL, CS, or CSKP] to allow for
        any orientation.

        The intersection calculation is based on an iterative procedure (250
        iterations maximum) and a tolerance of 1.0E-4.  The approximate
        location of a node should be sufficient to determine a unique
        intersection if more than one intersection point is possible.  Tangent
        "intersections" should be avoided.  If an intersection is not found,
        the node is placed at the last iteration location.

        This command is also valid in the /MAP processor.
        """
        command = f"MOVE,{node},{kc1},{x1},{y1},{z1},{kc2},{x2},{y2},{z2}"
        return self.run(command, **kwargs)

    def n(self, node="", x="", y="", z="", thxy="", thyz="", thzx="", **kwargs) -> int:
        """Define a node.

        APDL Command: N

        Defines a node in the active coordinate system [CSYS].  The
        nodal coordinate system is parallel to the global Cartesian
        system unless rotated.  Rotation angles are in degrees and
        redefine any previous rotation angles.  See the NMODIF, NANG,
        NROTAT, and NORA commands for other rotation options.

        Parameters
        ----------
        node
            Node number to be assigned.  A previously defined node of
            the same number will be redefined.  Defaults to the
            maximum node number used +1.

        x, y, z
            Node location in the active coordinate system (R, θ, Z for
            cylindrical, R, θ, Φ for spherical or toroidal).

        thxy
            First rotation about nodal Z (positive X toward Y).

        thyz
            Second rotation about nodal X (positive Y toward Z).

        thzx
            Third rotation about nodal Y (positive Z toward X).

        Returns
        -------
        int
            Node number of the generated node.

        Examples
        --------
        Create a node at ``(0, 1, 1)``

        >>> nnum = mapdl.n("", 0, 1, 1)
        >>> nnum
        1

        Create a node at ``(4, 5, 1)`` with a node ID of 10

        >>> nnum = mapdl.n(10, 4, 5, 1)
        >>> nnum
        10

        """
        command = f"N,{node},{x},{y},{z},{thxy},{thyz},{thzx}"
        msg = self.run(command, **kwargs)
        if msg:
            res = re.search(r"(NODE\s*)([0-9]+)", msg)
            if res is not None:
                return int(res.group(2))

    def naxis(self, action="", val="", **kwargs):
        """Generates nodes for general axisymmetric element sections.

        APDL Command: NAXIS

        Parameters
        ----------
        action
            Specifies one of the following command behaviors:

            GEN - Generates nodes around the axis of an axisymmetric section (default).

            CLEAR - Clears all nodes around the axis of an axisymmetric section.

            EFACET - Specifies the number of facets per edge between nodal planes and integration
                     planes in the circumferential direction to display using
                     PowerGraphics.  This option is only valid with /ESHAPE,1
                     and RSYS,SOLU commands.

        val
            Tolerance value or number of facets per edge:

            TOLER - When Action = GEN, the tolerance to use for merging the generated nodes around
                    the axis.

            NUM - When Action = EFACET, the number of facets per element edge for element plots:

            AUTO - Use program-chosen facets per edge (default).

            1 - Use 1 facet per edge (default for elements with 9, 10, 11, or 12 nodal planes).
                Shows nodal and integration planes only.

            2 - Use 2 facets per edge (default for elements with 5, 6, 7, or 8 nodal planes,
                and maximum for elements with 9, 10, 11, or 12 nodal planes).

            3 - Use 3 facets per edge (default for elements with 3 or 4 nodal planes, and
                maximum for elements with 6, 7, or 8 nodal planes).

            4 - Use 4 facets per edge (maximum for elements with 5 nodal planes).

            5 - Use 5 facets per edge (maximum for elements with 4 nodal planes).

            6 - Use 6 facets per edge (maximum for elements with 3 nodal planes).

        Notes
        -----
        The NAXIS command generates or clears the nodes for general
        axisymmetric element sections. The command applies to elements SURF159,
        SOLID272, and SOLID273.

        The generate option (Action = GEN) operates automatically on any
        current-technology axisymmetric element. Any nodes within the tolerance
        value (TOLER) of the axis are merged into a single node. The default
        tolerance is 1.0e-4.

        If you want to change the number of nodes, use the clear option (Action
        = CLEAR) before regenerating the nodes.

        To cause the 3-D element plot to appear more like the actual 3-D model,
        use NAXIS,EFACET,NUM, where NUM > 1. In this case, the coordinate
        system specified for displaying element and nodal results (RSYS) must
        be solution (RSYS,SOLU); otherwise, ANSYS resets NUM to 1.
        """
        command = f"NAXIS,{action},{val}"
        return self.run(command, **kwargs)

    def nang(
        self,
        node="",
        x1="",
        x2="",
        x3="",
        y1="",
        y2="",
        y3="",
        z1="",
        z2="",
        z3="",
        **kwargs,
    ):
        """Rotates a nodal coordinate system by direction cosines.

        APDL Command: NANG

        Parameters
        ----------
        node
            Rotate coordinate system of this node.

        x1, x2, x3
            Global X, Y, Z components of a unit vector in new nodal X
            direction.

        y1, y2, y3
            Global X, Y, Z components of a unit vector in new nodal Y
            direction.

        z1, z2, z3
            Global X, Y, Z components of a unit vector in new nodal Z
            direction.

        Notes
        -----
        Rotates a nodal coordinate system to the orientation specified by the
        X, Y and Z direction cosines.  Existing rotation specifications on the
        node are redefined.  If only two of the three unit vectors are
        specified, the third is defined according to the right hand rule.  It
        is the responsibility of the user to ensure that input direction
        cosines are orthogonal in a right-handed system.

        See the NMODIF, NROTAT, and NORA commands for other rotation options.
        """
        command = f"NANG,{node},{x1},{x2},{x3},{y1},{y2},{y3},{z1},{z2},{z3}"
        return self.run(command, **kwargs)

    def ndele(self, node1="", node2="", ninc="", **kwargs):
        """Deletes nodes.

        APDL Command: NDELE

        Parameters
        ----------
        node1, node2, ninc
            Delete nodes from NODE1 to NODE2 (defaults to NODE1) in steps of
            NINC (defaults to 1).  If NODE1 = ALL, NODE2 and NINC are ignored
            and all selected nodes [NSEL] are deleted.  If NODE1 = P, graphical
            picking is enabled and all remaining command fields are ignored
            (valid only in the GUI).  A component name may also be substituted
            for NODE1.

        Notes
        -----
        Deletes selected nodes that are not connected to elements.  Nodes may
        also be redefined instead of deleted, if desired.  Boundary conditions
        (displacements, forces, etc.) as well as any coupling or constraint
        equations containing the deleted nodes are also deleted.

        This command is also valid in the /MAP processor.
        """
        command = f"NDELE,{node1},{node2},{ninc}"
        return self.run(command, **kwargs)

    def ndist(self, nd1="", nd2="", **kwargs):
        """Calculates and lists the distance between two nodes.

        APDL Command: NDIST

        Parameters
        ----------
        nd1
            First node in distance calculation.  If ND1 = P, graphical picking
            is enabled and all remaining command fields are ignored (valid only
            in the GUI).
        nd2
            Second node in distance calculation.

        Returns
        -------
        list
            ``[DIST, X, Y, Z]`` distance between two nodes.

        Notes
        -----
        NDIST lists the distance between nodes ND1 and ND2, as well as the
        current coordinate system offsets from ND1 to ND2, where the X, Y, and
        Z locations of ND1 are subtracted from the X, Y, and Z locations of ND2
        (respectively) to determine the offsets.  NDIST is valid in any
        coordinate system except toroidal [CSYS,3].
        NDIST returns a variable, called "_RETURN," which contains the distance
        value. You can use this value for various purposes, such as the
        calculation of distributed loads. In interactive mode, you can access
        this command by using the Model Query Picker (Utility Menu> List>
        Picked Entities), where you can also access automatic annotation
        functions and display the value on your model.
        This command is valid in any processor.

        Examples
        --------
        Compute the distance between two nodes.

        >>> node1 = (0, 8, -3)
        >>> node2 = (13, 5, 7)
        >>> node_num1 = mapdl.n("", *node1)
        >>> node_num2 = mapdl.n("", *node2)
        >>> node_dist = mapdl.ndist(node_num1, node_num2)
        >>> node_dist
        [16.673332000533065, 13.0, -3.0, 10.0]

        """

        return parse.parse_ndist(self.run(f"NDIST,{nd1},{nd2}", **kwargs))

    def ngen(
        self,
        itime="",
        inc="",
        node1="",
        node2="",
        ninc="",
        dx="",
        dy="",
        dz="",
        space="",
        **kwargs,
    ):
        """Generates additional nodes from a pattern of nodes.

        APDL Command: NGEN

        Parameters
        ----------
        itime, inc
            Do this generation operation a total of ITIME times, incrementing
            all nodes in the given pattern by INC each time after the first.
            ITIME must be > 1 for generation to occur.

        node1, node2, ninc
            Generate nodes from the pattern of nodes beginning with NODE1 to
            NODE2 (defaults to NODE1) in steps of NINC (defaults to 1).  If
            NODE1 = ALL, NODE2 and NINC are ignored and the pattern is all
            selected nodes [NSEL].  If NODE1 = P, graphical picking is enabled
            and all remaining command fields are ignored (valid only in the
            GUI).  A component name may also be substituted for NODE1 (NODE2
            and NINC are ignored).

        dx, dy, dz
            Node location increments in the active coordinate system (DR, Dθ,
            DZ for cylindrical, DR, Dθ, DΦ for spherical or toroidal).

        space
            Spacing ratio.  Ratio of last division size to first division size.
            If > 1.0, divisions increase.  If < 1.0, divisions decrease.  Ratio
            defaults to 1.0 (uniform spacing).

        Notes
        -----
        Generates additional nodes from a given node pattern.  Generation is
        done in the active coordinate system.  Nodes in the pattern may have
        been generated in any coordinate system.

        This command is also valid in the /MAP processor.
        """
        command = f"NGEN,{itime},{inc},{node1},{node2},{ninc},{dx},{dy},{dz},{space}"
        return self.run(command, **kwargs)

    def nkpt(self, node="", npt="", **kwargs):
        """Defines a node at an existing keypoint location.

        APDL Command: NKPT

        Parameters
        ----------
        node
            Arbitrary reference number for node.  If zero or blank, defaults to
            the highest node number +1 [NUMSTR].

        npt
            Keypoint number defining global X, Y, Z  location.  If NPT = All,
            then a node will be placed at each selected keypoint.  If NPT = P,
            graphical picking is enabled and all remaining command fields are
            ignored (valid only in the GUI).  A component name may also be
            substituted for NPT.
        """
        command = f"NKPT,{node},{npt}"
        return self.run(command, **kwargs)

    def nlist(
        self,
        node1="",
        node2="",
        ninc="",
        lcoord="",
        sort1="",
        sort2="",
        sort3="",
        kinternal="",
        **kwargs,
    ):
        """Lists nodes.

        APDL Command: NLIST

        Parameters
        ----------
        node1, node2, ninc
            List nodes from NODE1 to NODE2 (defaults to NODE1) in steps of NINC
            (defaults to 1).  If NODE1 = ALL (default), NODE2 and NINC are
            ignored and all selected nodes [NSEL] are listed.  If NODE1 = P,
            graphical picking is enabled and all remaining command fields are
            ignored (valid only in the GUI).  A component name may also be
            substituted for NODE1 (NODE2 and NINC are ignored).

        lcoord
            Coordinate listing key:

            (blank) - List all nodal information

            COORD - Suppress all but the XYZ coordinates (shown to a higher degree of accuracy than
                    when displayed with all information).

        sort1
            First item on which to sort.  Valid item names are NODE, X, Y, Z,
            THXY, THYZ, THXZ

        sort2, sort3
            Second and third items on which to sort.  Valid item names are the
            same as for SORT1.

        kinternal
            Internal nodes listing key:

            (blank) - List only external nodes.

            INTERNAL - List all nodes, including internal nodes.

        Notes
        -----
        Lists nodes in the active display coordinate system [DSYS].  Nodal
        coordinate rotation angles are also listed (relative to the global
        Cartesian coordinate system).

        Node listing can be in a sorted order (ascending).  SORT2, for example,
        will be carried out on nodes having equal values of SORT1.

        This command is valid in any processor.
        """
        command = (
            f"NLIST,{node1},{node2},{ninc},{lcoord},{sort1},{sort2},{sort3},{kinternal}"
        )
        return self.run(command, **kwargs)

    def nmodif(self, node="", x="", y="", z="", thxy="", thyz="", thzx="", **kwargs):
        """Modifies an existing node.

        APDL Command: NMODIF

        Parameters
        ----------
        node
            Modify coordinates of this node.  If ALL, modify coordinates of all
            selected nodes [NSEL].  If NODE = P, graphical picking is enabled
            and all remaining command fields are ignored (valid only in the
            GUI).  A component name may also be substituted for NODE.

        x, y, z
            Replace the previous coordinate values assigned to this node with
            these corresponding coordinate values.  Values are interpreted in
            the active coordinate system (R, θ, Z for cylindrical; R, θ, Φ for
            spherical or toroidal).  Leaving any of these fields blank retains
            the previous value(s).

        thxy
            First rotation of nodal coordinate system about nodal Z (positive X
            toward Y). Leaving this field blank retains the previous value.

        thyz
            Second rotation of nodal coordinate system about nodal X (positive
            Y toward Z). Leaving this field blank retains the previous value.

        thzx
            Third rotation of nodal coordinate system about nodal Y (positive Z
            toward X). Leaving this field blank retains the previous value.

        Notes
        -----
        Modifies an existing node.  Nodal coordinate system rotation angles are
        in degrees and redefine any existing rotation angles.  Nodes can also
        be redefined with the N command.

        See the NROTAT, NANG, and NORA commands for other rotation options.

        This command is also valid in the /MAP processor.
        """
        command = f"NMODIF,{node},{x},{y},{z},{thxy},{thyz},{thzx}"
        return self.run(command, **kwargs)

    def nora(self, area="", ndir="", **kwargs):
        """Rotates nodal coordinate systems to surface normal

        APDL Command: NORA

        Parameters
        ----------
        area
            The area number containing the nodes to be rotated to their
            normals. If ALL, applies to all selected areas (see the ASEL
            command). If AREA = P, graphical picking is enabled.

        ndir
            Direction of the normal. If NDIR = -1, the nodal coordinate system
            is rotated in the opposite direction of the surface normal. The
            default is the same direction as the surface normal.

        Notes
        -----
        The NORA command rotates the X-axis of the nodal coordinate system to
        the surface normal. The rotated nodal coordinate systems may be
        displayed through the /PSYMB command. In case multiple areas are
        selected, there could be conflicts at the boundaries. If a node belongs
        to two areas that have a different normal, its nodal coordinate system
        will be rotated to the area normal with the lowest number. You can use
        the AREVERSE and ANORM commands to rotate the surface normals in the
        appropriate direction. Keep the following in mind when using the NORA
        command:

        If the nodal coordinate system is parallel to the global Cartesian
        system, it is not displayed through the /PSYMB command.

        Previously specified rotation on the selected nodes are overridden.
        """
        command = f"NORA,{area},{ndir}"
        return self.run(command, **kwargs)

    def norl(self, line="", area="", ndir="", **kwargs):
        """APDL Command: NORL

        Rotates nodal coordinate systems perpendicular to line normal

        Parameters
        ----------
        line
            Line number containing the nodes to be rotated. If ALL, applies to
            all selected lines (see the LSEL command). If LINE = P, graphical
            picking is enabled.

        area
            The area number containing the selected lines. The normal of the
            line(s) selected is supposed to lie on this area. Defaults to the
            lowest numbered selected area containing the line number.

        ndir
            Direction of the normal. If NDIR = -1, the nodal coordinate system
            is rotated in the opposite direction of the line normal. The
            default is the same direction as the surface normal.

        Notes
        -----
        The NORL command rotates the X-axis of the nodal coordinate
        perpendicular to the line normal. The rotated nodal coordinate systems
        may be displayed through the /PSYMB command. In case multiple lines are
        selected, there could be conflicts at the boundaries. If a node belongs
        to two lines that have a different normal, its nodal coordinate system
        will be rotated to the line normal with the lowest number. Keep the
        following in mind when using the NORL command:

        If the nodal coordinate system is parallel to the global Cartesian
        system, it is not displayed through the /PSYMB command.

        Previously specified rotation on the selected nodes are overridden.
        """
        command = f"NORL,{line},{area},{ndir}"
        return self.run(command, **kwargs)

    def nplot(self, knum="", **kwargs):
        """Displays nodes.

        APDL Command: NPLOT

        Parameters
        ----------
        knum
            Node number key:

            0 - No node numbers on display.

            1 - Include node numbers on display.  See also /PNUM command.

        Notes
        -----
        Produces a node display.  Only selected nodes [NSEL] are displayed.
        Elements need not be defined.  See the DSYS command for display
        coordinate system.

        This command is valid in any processor.
        """
        command = f"NPLOT,{knum}"
        return self.run(command, **kwargs)

    def nread(self, fname="", ext="", **kwargs):
        """Reads nodes from a file.

        APDL Command: NREAD

        Parameters
        ----------
        fname
            File name and directory path (248 characters maximum, including the
            characters needed for the directory path).  An unspecified
            directory path defaults to the working directory; in this case, you
            can use all 248 characters for the file name.

        ext
            Filename extension (eight-character maximum).

        Notes
        -----
        The read operation is not necessary in a standard ANSYS run but is
        provided as a convenience to users wanting to read a coded node file,
        such as from another mesh generator or from a CAD/CAM program.  Data
        should be formatted as produced with the NWRITE command. Only nodes
        that are within the node range specified with the NRRANG command are
        read from the file.  Duplicate nodes already in the database will be
        overwritten.  The file is rewound before and after reading.  Reading
        continues until the end of the file.
        """
        command = f"NREAD,{fname},{ext}"
        return self.run(command, **kwargs)

    def nrotat(self, node1="", node2="", ninc="", **kwargs):
        """Rotates nodal coordinate systems into the active system.

        APDL Command: NROTAT

        Parameters
        ----------
        node1, node2, ninc
            Rotate nodes from NODE1 to NODE2 (defaults to NODE1) in steps of
            NINC (defaults to 1).  If NODE1 = ALL, NODE2 and NINC are ignored
            and all selected nodes [NSEL] are rotated.  If NODE1 = P, graphical
            picking is enabled and all remaining command fields are ignored
            (valid only in the GUI).  A component name may also be substituted
            for NODE1 (NODE2 and NINC are ignored).

        Notes
        -----
        Rotates nodal coordinate systems into the active coordinate system.
        Nodal coordinate systems may be automatically rotated into the active
        (global or local) coordinate system as follows:  Rotations in Cartesian
        systems will have nodal x directions rotated parallel to the Cartesian
        X direction.  Rotations in cylindrical, spherical or toroidal systems
        will have the nodal x directions rotated parallel to the R direction.
        Nodes at (or near) a zero radius location should not be rotated.  Nodal
        coordinate directions may be displayed [/PSYMB].  Nodal forces and
        constraints will also appear rotated when displayed if the nodal
        coordinate system is rotated.

        ANSYS LS-DYNA (explicit dynamics) does not support the NROTAT command.
        If you have rotated nodes in the implicit phase of an implicit-to-
        explicit sequential solution, you must rotate the nodes back to the
        global Cartesian direction before switching from implicit to explicit
        elements (ETCHG,ITE). Use the EDNROT command in the explicit run to
        maintain the same displacement constraints as were used on rotated
        nodes in the implicit run.

        Note:: : When the  nodal coordinate systems are defined, they remain
        parallel to the global Cartesian system unless subsequently rotated.

        Previously specified rotations on the specified nodes are overridden.

        See the NMODIF, NANG, and NORA commands for other rotation options.
        """
        command = f"NROTAT,{node1},{node2},{ninc}"
        return self.run(command, **kwargs)

    def nrrang(self, nmin="", nmax="", ninc="", **kwargs):
        """Specifies the range of nodes to be read from the node file.

        APDL Command: NRRANG

        Parameters
        ----------
        nmin, nmax, ninc
            Node range is defined from NMIN (defaults to 1) to NMAX (defaults
            to 99999999) in steps of NINC (defaults to 1).

        Notes
        -----
        Defines the range of nodes to be read [NREAD] from the node file.  Also
        implies an element range since only elements fully attached to these
        nodes will be read from the element file.
        """
        command = f"NRRANG,{nmin},{nmax},{ninc}"
        return self.run(command, **kwargs)

    def nscale(
        self, inc="", node1="", node2="", ninc="", rx="", ry="", rz="", **kwargs
    ):
        """Generates a scaled set of nodes from a pattern of nodes.

        APDL Command: NSCALE

        Parameters
        ----------
        inc
            Do this scaling operation one time, incrementing all nodes in the
            given pattern by INC.  If INC = 0, nodes will be redefined at the
            scaled locations.

        node1, node2, ninc
            Scale nodes from pattern of nodes beginning with NODE1 to NODE2
            (defaults to NODE1) in steps of NINC (defaults to 1).  If  NODE1 =
            ALL, NODE2 and NINC are ignored and pattern is all selected nodes
            [NSEL].  If NODE1 = P, graphical picking is enabled and all
            remaining command fields are ignored (valid only in the GUI).  A
            component name may also be substituted for NODE1 (NODE2 and NINC
            are ignored).

        rx, ry, rz
            Scale factor ratios.  Scaling is relative to the origin of the
            active coordinate system (RR, Rθ, RZ for cylindrical, RR, Rθ, RΦ
            for spherical or toroidal).  If absolute value of ratio > 1.0,
            pattern is enlarged.  If < 1.0, pattern is reduced.  Ratios default
            to 1.0 (each).

        Notes
        -----
        Generates a scaled pattern of nodes from a given node pattern.  Scaling
        is done in the active coordinate system.  Nodes in the pattern may have
        been generated in any coordinate system.

        This command is also valid in the /MAP processor.
        """
        command = f"NSCALE,{inc},{node1},{node2},{ninc},{rx},{ry},{rz}"
        return self.run(command, **kwargs)

    def nsmooth(self, npass="", **kwargs):
        """Smooths selected nodes among selected elements.

        APDL Command: NSMOOTH

        Parameters
        ----------
        npass
            Number of smoothing passes. Defaults to 3.

        Notes
        -----
        Repositions each selected node at the average position of its immediate
        neighbors on the selected elements. The node positions converge after
        some number of smoothing passes. For some initial conditions, NPASS may
        need to be much larger than 3. If the boundary of a mesh is to be
        undisturbed (usually desirable), the boundary nodes should be
        unselected before issuing NSMOOTH.
        """
        command = f"NSMOOTH,{npass}"
        return self.run(command, **kwargs)

    def nsym(self, ncomp="", inc="", node1="", node2="", ninc="", **kwargs):
        """Generates a reflected set of nodes.

        APDL Command: NSYM

        Parameters
        ----------
        ncomp
            Symmetry key:

            X - X (or R) symmetry (default).

            Y - Y (or θ) symmetry.

            Z - Z (or Φ) symmetry.

        inc
            Increment all nodes in the given pattern by INC to form the
            reflected node pattern.

        node1, node2, ninc
            Reflect nodes from pattern beginning with NODE1 to NODE2 (defaults
            to NODE1) in steps of NINC (defaults to 1).  If  NODE1 = ALL, NODE2
            and NINC are ignored and pattern is all selected nodes [NSEL].  If
            NODE1 = P, graphical picking is enabled and all remaining command
            fields are ignored (valid only in the GUI).  A component name may
            also be substituted for NODE1 (NODE2 and NINC are ignored).

        Notes
        -----
        Generates nodes from a given node pattern by a symmetry reflection.
        Reflection is done in the active coordinate system by changing a
        particular coordinate sign.  Nodes in the pattern may have been
        generated in any coordinate system.   Nodal rotation angles are not
        reflected.

        Symmetry reflection may be used with any node pattern, in any
        coordinate system, as many times as desired.  Reflection is
        accomplished by a coordinate sign change (in the active coordinate
        system).  For example, an X-reflection in a Cartesian coordinate system
        generates additional nodes from a given pattern, with a node increment
        added to each node number, and an X coordinate sign change.  An
        R-reflection in a cylindrical coordinate system gives a reflected
        "radial" location by changing the "equivalent" Cartesian (i.e., the
        Cartesian system with the same origin as the active cylindrical system)
        X and Y coordinate signs.  An R-reflection in a spherical coordinate
        system gives a reflected "radial" location by changing the equivalent
        Cartesian X, Y, and Z coordinate location signs.  Nodal coordinate
        system rotation angles are not reflected.
        """
        command = f"NSYM,{ncomp},{inc},{node1},{node2},{ninc}"
        return self.run(command, **kwargs)

    def nwrite(self, fname="", ext="", kappnd="", **kwargs):
        """Writes nodes to a file.

        APDL Command: NWRITE

        Parameters
        ----------
        fname
            File name and directory path (248 characters maximum,
            including the characters needed for the directory path).
            An unspecified directory path defaults to the working
            directory; in this case, you can use all 248 characters
            for the file name.

        ext
            Filename extension (eight-character maximum).

        kappnd
            Append key:

            0 - Rewind file before the write operation.

            1 - Append data to the end of the existing file.

        Notes
        -----
        Writes selected nodes [NSEL] to a file.  The write operation
        is not necessary in a standard ANSYS run but is provided as a
        convenience to users wanting a coded node file.  Data are
        written in a coded format.  The format used is (I8, 6G20.13)
        to write out NODE,X,Y,Z,THXY,THYZ,THZX.  If the last number is
        zero (i.e., THZX = 0), or the last set of numbers are zero,
        they are not written but are left blank. Therefore, you must
        use a formatted read to process this file.  Coordinate values
        are in the global Cartesian system.
        """
        return self.run(f"NWRITE,{fname},{ext},,{kappnd}", **kwargs)

    def quad(
        self,
        node1="",
        nintr="",
        node2="",
        nfill="",
        nstrt="",
        ninc="",
        pkfac="",
        **kwargs,
    ):
        """Generates a quadratic line of nodes from three nodes.

        APDL Command: QUAD

        Parameters
        ----------
        node1
            Begin fill-in from this node location.  If NODE1 = P, graphical
            picking is enabled and all remaining command fields are ignored
            (valid only in the GUI).

        nintr
            Intermediate or guiding node.  Quadratic curve will pass through
            this location.  NINTR may have any node number and any location.
            If the quadratic line also generates a node with number NINTR, the
            generated location overrides the previous NINTR location.

        node2
            End quadratic fill-in at this node location.

        nfill
            Fill-in NFILL nodes between NODE1 and NODE2 (defaults to
            ``|NODE2-NODE1|-1``).  NFILL must be positive.

        nstrt
            Node number assigned to first filled-in node (defaults to NODE1 +
            NINC).

        ninc
            Add this increment to each of the remaining filled-in node numbers
            (may be positive or negative).  Defaults to ``(NODE2-NODE1)/(NFILL +
            1)``, i.e., linear interpolation.

        pkfac
            Peak location factor.  If PKFAC=0.5, the peak of the quadratic
            shape occurs at the NINTR location.  If 0.0 < PKFAC < 0.5, the peak
            occurs to the NODE2 side of the NINTR location.  If 0.5 < PKFAC <
            1.0, the peak occurs to the NODE1 side of the NINTR location.
            Defaults to 0.5.

        Notes
        -----
        Generates a quadratic line of nodes (in the active coordinate system)
        from three nodes.  The three nodes determine the plane of the curve and
        may have been defined in any coordinate system.  Any number of nodes
        may be filled-in and any node number sequence may be assigned.

        The quadratic line feature uses three nodes (NODE1,NINTR,NODE2) to
        determine the plane of the curve.  The curve passes through the three
        points, beginning from NODE1, through the intermediate (or guiding)
        point NINTR, and toward NODE2.

        Generated nodes are also quadratically spaced.  If the guiding node
        number is within the set being generated, it will be relocated
        according to the quadratic spacing.

        The peak location factor is used to determine how the quadratic fits
        through the three points.  Various nodal progressions can be obtained
        by different combinations of PKFAC and the guiding node location.  If
        the guiding node is at mid-length between NODE1 and NODE2, 0.293:
        PKFAC< 0.707 will ensure that all generated nodes fall within the
        NODE1,NODE2 bounds.  In the limit, as PKFAC approaches 0.0, the peak
        approaches the line through NODE1 and NINTR at an infinite distance
        from NODE1.  The QUAD command generates quadratic lines of nodes, which
        in turn may be used as a base line for generating irregular surfaces of
        nodes (by repeating [``*REPEAT``], generating [NGEN, NSCALE], etc.).
        Irregular surfaces may also be generated with the meshing commands.
        """
        command = f"QUAD,{node1},{nintr},{node2},{nfill},{nstrt},{ninc},{pkfac}"
        return self.run(command, **kwargs)

    def transfer(self, kcnto="", inc="", node1="", node2="", ninc="", **kwargs):
        """Transfers a pattern of nodes to another coordinate system.

        APDL Command: TRANSFER

        Parameters
        ----------
        kcnto
            Reference number of coordinate system where the pattern is to be
            transferred.  Transfer occurs from the active coordinate system.

        inc
            Increment all nodes in the given pattern by INC to form the
            transferred node pattern.

        node1, node2, ninc
            Transfer nodes from pattern beginning with NODE1 to NODE2 (defaults
            to NODE1) in steps of NINC (defaults to 1).  If NODE1 = ALL, NODE2
            and NINC are ignored and the pattern is all selected nodes [NSEL].
            If NODE1 = P, graphical picking is enabled and all remaining
            command fields are ignored (valid only in the GUI).  A component
            may be substituted for NODE1 (NODE2 and NINC are ignored).

        Notes
        -----
        Transfers a pattern of nodes from one coordinate system to another.
        Coordinate systems may be translated and rotated relative to each
        other.  Initial pattern may be generated in any coordinate system.
        Coordinate values are interpreted in the active coordinate system and
        are transferred directly.

        A model generated in one coordinate system may be transferred to
        another coordinate system.  The user may define several coordinate
        systems (translated and rotated from each other), generate a model in
        one coordinate system, and then repeatedly transfer the model to other
        coordinate systems.  The model may be generated in any type of
        coordinate system (Cartesian, cylindrical, etc.) and transferred to any
        other type of coordinate system.  Coordinate values (X, Y, Z, or R,: θ,
        Z, or etc.) of the model being transferred are interpreted in the
        active coordinate system type, regardless of how they were generated.
        Values are transferred directly and are interpreted according to the
        type of coordinate system being transferred to.  For example,
        transferring from a Cartesian coordinate system to a cylindrical
        coordinate system (not recommended) would cause X = 2.0 and Y = 3.0
        values to be directly interpreted as R = 2.0 and θ = 3.0 values,
        respectively.

        This command is also valid in the /MAP processor.
        """
        command = f"TRANSFER,{kcnto},{inc},{node1},{node2},{ninc}"
        return self.run(command, **kwargs)



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from ansys.mapdl.core._commands import parse


class Booleans:
    def aadd(
        self,
        na1="",
        na2="",
        na3="",
        na4="",
        na5="",
        na6="",
        na7="",
        na8="",
        na9="",
        **kwargs,
    ):
        """Adds separate areas to create a single area.

        APDL Command: AADD

        Parameters
        ----------
        na1, na2, na3, . . . , na9
            Numbers of areas to be added.  If NA1 = ALL, add all selected areas
            and ignore NA2 to NA9.  If NA1 = P, graphical picking is enabled
            and all remaining arguments are ignored (valid only in the GUI).  A
            component name may also be substituted for NA1.

        Notes
        -----
        The areas must be coplanar.  The original areas (and their
        corresponding lines and keypoints) will be deleted by default.
        See the BOPTN command for the options available to Boolean
        operations.  Element attributes and solid model boundary
        conditions assigned to the original entities will not be
        transferred to the new entities generated.  Concatenated entities
        are not valid with this command.

        Examples
        --------
        Generate two areas and combine them.

        >>> a1 = mapdl.rectng(2.5, 3.5, 0, 10)
        >>> a2 = mapdl.cyl4(0, 10, 2.5, 0, 3.5, 90)
        >>> a_comb = mapdl.aadd(a1, a2)
        >>> a_comb
        3

        """
        command = f"AADD,{na1},{na2},{na3},{na4},{na5},{na6},{na7},{na8},{na9}"
        return parse.parse_output_areas(self.run(command, **kwargs))

    def aglue(
        self,
        na1="",
        na2="",
        na3="",
        na4="",
        na5="",
        na6="",
        na7="",
        na8="",
        na9="",
        **kwargs,
    ):
        """Generates new areas by "gluing" areas.

        APDL Command: AGLUE

        Parameters
        ----------
        na1, na2, na3, . . . , na9
            Numbers of the areas to be glued.  If NA1 = ALL, all selected areas
            will be glued (NA2 to NA9 will be ignored).  If NA1 = P, graphical
            picking is enabled and all remaining arguments are ignored (valid
            only in the GUI).  A component name may also be substituted for
            NA1.

        Notes
        -----
        Use of the AGLUE command generates new areas by "gluing" input areas.
        The glue operation redefines the input areas so that they share lines
        along their common boundaries.  The new areas encompass the same
        geometry as the original areas.  This operation is only valid if the
        intersection of the input areas are lines along the boundaries of those
        areas.  See the Modeling and Meshing Guide for an illustration.  See
        the BOPTN command for an explanation of the options available to
        Boolean operations.  Element attributes and solid model boundary
        conditions assigned to the original entities will not be transferred to
        new entities generated.

        The AGLUE command results in the merging of lines and keypoints at the
        common area boundaries. The lines and keypoints of the lower numbered
        area will be kept. This means one must be aware of area numbering when
        multiple AGLUE commands are applied to avoid any "ungluing" of
        geometry.
        """
        command = f"AGLUE,{na1},{na2},{na3},{na4},{na5},{na6},{na7},{na8},{na9}"
        return self.run(command, **kwargs)

    def aina(
        self,
        na1="",
        na2="",
        na3="",
        na4="",
        na5="",
        na6="",
        na7="",
        na8="",
        na9="",
        **kwargs,
    ):
        """Finds the intersection of areas.

        APDL Command: AINA

        Parameters
        ----------
        na1, na2, na3, . . . , na9
            Numbers of areas to be intersected.  If NA1 = ALL, NA2 to NA9 are
            ignored and the intersection of all selected areas is found.  If
            NA1 = P, graphical picking is enabled and all remaining arguments
            are ignored (valid only in the GUI).  A component name may also be
            substituted for NA1.

        Notes
        -----
        Finds the common (not pairwise) intersection of areas.  The common
        intersection is defined as the regions shared (in common) by all areas
        listed on this command.  New areas will be generated where the original
        areas intersect.  If the regions of intersection are only lines, new
        lines will be generated instead.  See the Modeling and Meshing Guide
        for an illustration.  See the BOPTN command for the options available
        to Boolean operations.  Element attributes and solid model boundary
        conditions assigned to the original entities will not be transferred to
        the new entities generated.
        """
        command = f"AINA,{na1},{na2},{na3},{na4},{na5},{na6},{na7},{na8},{na9}"
        return self.run(command, **kwargs)

    def ainp(
        self,
        na1="",
        na2="",
        na3="",
        na4="",
        na5="",
        na6="",
        na7="",
        na8="",
        na9="",
        **kwargs,
    ):
        """Finds the pairwise intersection of areas.

        APDL Command: AINP

        Parameters
        ----------
        na1, na2, na3, . . . , na9
            Numbers of areas to be intersected pairwise.  If NA1 = ALL, NA2 to
            NA9 are ignored and the pairwise intersection of all selected areas
            is found.  If NA1 = P, graphical picking is enabled and all
            remaining arguments are ignored (valid only in the GUI).  A
            component name may be substituted for NA1.

        Notes
        -----
        Finds the pairwise intersection of areas.  The pairwise intersection is
        defined as all regions shared by any two or more areas listed on this
        command.  New areas will be generated where the original areas
        intersect pairwise.  If the regions of pairwise intersection are only
        lines, new lines will be generated.  See the Modeling and Meshing Guide
        for an illustration.  See the BOPTN command for the options available
        to Boolean operations.  Element attributes and solid model boundary
        conditions assigned to the original entities will not be transferred to
        the new entities generated.
        """
        command = f"AINP,{na1},{na2},{na3},{na4},{na5},{na6},{na7},{na8},{na9}"
        return self.run(command, **kwargs)

    def ainv(self, na="", nv="", **kwargs):
        """Finds the intersection of an area with a volume.

        APDL Command: AINV

        Parameters
        ----------
        na
            Number of area to be intersected.  If P, graphical picking is
            enabled and all remaining arguments are ignored (valid only in the
            GUI).

        nv
            Number of volume to be intersected.

        Notes
        -----
        New areas will be generated where the areas intersect the volumes.  If
        the regions of intersection are only lines, new lines will be generated
        instead.  See the Modeling and Meshing Guide for an illustration.  See
        the BOPTN command for the options available to Boolean operations.
        Element attributes and solid model boundary conditions assigned to the
        original entities will not be transferred to the new entities
        generated.
        """
        command = f"AINV,{na},{nv}"
        return self.run(command, **kwargs)

    def aovlap(
        self,
        na1="",
        na2="",
        na3="",
        na4="",
        na5="",
        na6="",
        na7="",
        na8="",
        na9="",
        **kwargs,
    ):
        """Overlaps areas.

        APDL Command: AOVLAP

        Parameters
        ----------
        na1, na2, na3, . . . , na9
            Numbers of areas to be operated on.  If NA1 = ALL, use all selected
            areas and ignore NA2 to NA9.  If NA1 = P, graphical picking is
            enabled and all remaining arguments are ignored (valid only in the
            GUI).  A component name may also be substituted for NA1.

        Notes
        -----
        Generates new areas which encompass the geometry of all the input
        areas. The new areas are defined by the regions of intersection of the
        input areas, and by the complementary (non-intersecting) regions.  See
        Solid Modeling in the Modeling and Meshing Guide for an illustration.
        This operation is only valid when the region of intersection is an
        area.  See the BOPTN command for an explanation of the options
        available to Boolean operations.  Element attributes and solid model
        boundary conditions assigned to the original entities will not be
        transferred to the new entities generated.
        """
        command = f"AOVLAP,{na1},{na2},{na3},{na4},{na5},{na6},{na7},{na8},{na9}"
        return self.run(command, **kwargs)

    def aptn(
        self,
        na1="",
        na2="",
        na3="",
        na4="",
        na5="",
        na6="",
        na7="",
        na8="",
        na9="",
        **kwargs,
    ):
        """Partitions areas.

        APDL Command: APTN

        Parameters
        ----------
        na1, na2, na3, . . . , na9
            Numbers of areas to be operated on.  If NA1 = ALL, NA2 to NA9 are
            ignored and all selected areas are used.  If NA1 = P, graphical
            picking is enabled and all remaining arguments are ignored (valid
            only in the GUI).  A component name may be substituted for NA1.

        Notes
        -----
        Partitions areas that intersect.  This command is similar to the
        combined functionality of the ASBA and AOVLAP commands.  If the
        intersection of two or more areas is an area (i.e., planar), new areas
        will be created with boundaries that conform to the area of
        intersection and to the boundaries of the non-intersecting portions of
        the input areas [AOVLAP].  If the intersection is a line (i.e., not
        planar), the areas will be subtracted, or divided, along the line(s) of
        intersection [ASBA].  Both types of intersection can occur during a
        single APTN operation.  Areas that do not intersect will not be
        modified.  See the Modeling and Meshing Guide for an illustration.  See
        the BOPTN command for an explanation of the options available to
        Boolean operations.  Element attributes and solid model boundary
        conditions assigned to the original entities will not be transferred to
        the new entities generated.
        """
        command = f"APTN,{na1},{na2},{na3},{na4},{na5},{na6},{na7},{na8},{na9}"
        return self.run(command, **kwargs)

    def asba(self, na1="", na2="", sepo="", keep1="", keep2="", **kwargs) -> int:
        """Subtracts areas from areas.

        APDL Command: ASBA

        Generates new areas by subtracting the regions common to both
        NA1 and NA2 areas (the intersection) from the NA1 areas.  The
        intersection can be an area(s) or line(s).  If the
        intersection is a line and SEPO is blank, the NA1 area is
        divided at the line and the resulting areas will be connected,
        sharing a common line where they touch.  If SEPO is set to
        SEPO, NA1 is divided into two unconnected areas with separate
        lines where they touch.  See Solid Modeling in the Modeling
        and Meshing Guide for an illustration.  See the BOPTN command
        for an explanation of the options available to Boolean
        operations.  Element attributes and solid model boundary
        conditions assigned to the original entities will not be
        transferred to the new entities generated.  ASBA,ALL,ALL will
        have no effect since all the areas (in NA1) will be
        unavailable as NA2 areas.

        Parameters
        ----------
        na1
            Area (or areas, if picking is used) to be subtracted from.
            If ALL, use all selected areas.  Areas specified in this
            argument are not available for use in the NA2 argument.  A
            component name may also be substituted for NA1.

        na2
            Area (or areas, if picking is used) to subtract.  If ALL,
            use all selected areas (except those included in the NA1
            argument).  A component name may also be substituted for
            NA2.

        sepo
            Behavior if the intersection of the NA1 areas and the NA2 areas is
            a line or lines:

            (blank) - The resulting areas will share line(s) where they touch.

            SEPO - The resulting areas will have separate, but
                   coincident line(s) where they touch.

        keep1
            Specifies whether NA1 areas are to be deleted:

            (blank) - Use the setting of KEEP on the BOPTN command.

            DELETE - Delete NA1 areas after ASBA operation (override
            BOPTN command settings).

            KEEP - Keep NA1 areas after ASBA operation (override BOPTN
            command settings).

        keep2
            Specifies whether NA2 areas are to be deleted:

            (blank) - Use the setting of KEEP on the BOPTN command.

            DELETE - Delete NA2 areas after ASBA operation (override
            BOPTN command settings).

            KEEP - Keep NA2 areas after ASBA operation (override BOPTN
            command settings).

        Returns
        -------
        int
            Area number of the new area (if applicable)

        Examples
        --------
        Subtract a 0.5 x 0.5 rectangle from a 1 x 1 rectangle.

        >>> anum0 = mapdl.blc4(0, 0, 1, 1)
        >>> anum1 = mapdl.blc4(0.25, 0.25, 0.5, 0.5)
        >>> aout = mapdl.asba(anum0, anum1)
        >>> aout
        3

        """
        command = f"ASBA,{na1},{na2},{sepo},{keep1},{keep2}"
        return parse.parse_output_volume_area(self.run(command, **kwargs))

    def asbl(self, na="", nl="", keepa="", keepl="", **kwargs):
        """Subtracts lines from areas.

        APDL Command: ASBL

        Parameters
        ----------
        na
            Area (or areas, if picking is used) to be subtracted from.  If ALL,
            use all selected areas.  If P, graphical picking is enabled (valid
            only in the GUI) and remaining fields are ignored.  A component
            name may also be substituted for NA.

        nl
            Line (or lines, if picking is used) to subtract.  If ALL, use all
            selected lines.    A component name may also be substituted for NL.

        keepa
            Specifies whether NA areas are to be deleted:

            (blank) - Use the setting of KEEP on the BOPTN command.

            DELETE - Delete NA areas after ASBL operation (override BOPTN command settings).

            KEEP - Keep NA areas after ASBL operation (override BOPTN command settings).

        keepl
            Specifies whether NL lines are to be deleted:

            (blank) - Use the setting of KEEP on the BOPTN command.

            DELETE - Delete NL lines after ASBL operation (override BOPTN command settings).

            KEEP - Keep NL lines after ASBL operation (override BOPTN command settings).

        Notes
        -----
        Generates new areas by subtracting the regions common to both the areas
        and lines (the intersection) from the NA areas.  The intersection will
        be a line(s).  See Solid Modeling in the Modeling and Meshing Guide for
        an illustration.  See the BOPTN command for an explanation of the
        options available to Boolean operations.  Element attributes and solid
        model boundary conditions assigned to the original entities will not be
        transferred to the new entities generated.
        """
        command = f"ASBL,{na},{nl},,{keepa},{keepl}"
        return self.run(command, **kwargs)

    def asbv(self, na="", nv="", sepo="", keepa="", keepv="", **kwargs):
        """Subtracts volumes from areas.

        APDL Command: ASBV

        Parameters
        ----------
        na
            Area (or areas, if picking is used) to be subtracted from.  If ALL,
            use all selected areas.  If P, graphical picking is enabled (valid
            only in the GUI) and remaining fields are ignored.  A component
            name may also be substituted for NA.

        nv
            Volume (or volumes, if picking is used) to subtract.  If ALL, use
            all selected volumes.  A component name may also be substituted for
            NV.

        sepo
            Behavior if the intersection of the areas and the volumes is a line
            or lines:

            (blank) - The resulting areas will share line(s) where they touch.

            SEPO - The resulting areas will have separate, but coincident line(s) where they
                   touch.

        keepa
            Specifies whether NA areas are to be deleted:

            (blank) - Use the setting of KEEP on the BOPTN command.

            DELETE - Delete NA areas after ASBV operation (override BOPTN command settings).

            KEEP - Keep NA areas after ASBV operation (override BOPTN command settings).

        keepv
            Specifies whether NV volumes are to be deleted:

            (blank) - Use the setting of KEEP on the BOPTN command.

            DELETE - Delete volumes after ASBV operation (override BOPTN command settings).

            KEEP - Keep volumes after ASBV operation (override BOPTN command settings).

        Notes
        -----
        Generates new areas by subtracting the regions common to both NA areas
        and NV volumes (the intersection) from the NA areas.  The intersection
        can be an area(s) or line(s).  If the intersection is a line and SEPO
        is blank, the NA area is divided at the line and the resulting areas
        will be connected, sharing a common line where they touch.  If SEPO is
        set to SEPO, NA is divided into two unconnected areas with separate
        lines where they touch.  See Solid Modeling in the Modeling and Meshing
        Guide for an illustration.  See the BOPTN command for an explanation of
        the options available to Boolean operations.  Element attributes and
        solid model boundary conditions assigned to the original entities will
        not be transferred to the new entities generated.
        """
        command = f"ASBV,{na},{nv},{sepo},{keepa},{keepv}"
        return self.run(command, **kwargs)

    def asbw(self, na="", sepo="", keep="", **kwargs):
        """Subtracts the intersection of the working plane from areas (divides

        APDL Command: ASBW
        areas).

        Parameters
        ----------
        na
            Area (or areas, if picking is used) to be subtracted from.  If NA =
            ALL, use all selected areas.  If NA = P, graphical picking is
            enabled (valid only in the GUI).  A component name may also be
            input for NA.

        sepo
            Behavior of the created boundary.

            (blank) - The resulting areas will share line(s) where they touch.

            SEPO - The resulting areas will have separate, but coincident line(s).

        keep
            Specifies whether NA areas are to be deleted.

            (blank) - Use the setting of KEEP on the BOPTN command.

            DELETE - Delete NA areas after ASBW operation (override BOPTN command settings).

            KEEP - Keep NA areas after ASBW operation (override BOPTN command settings).

        Notes
        -----
        Generates new areas by subtracting the intersection of the working
        plane from the NA areas.  The intersection will be a line(s).  The
        working plane must not be in the same plane as the NA areas(s).  If
        SEPO is blank, the NA area is divided at the line and the resulting
        areas will be connected, sharing a common line where they touch.  If
        SEPO is set to SEPO, NA is divided into two unconnected areas with
        separate lines.  The SEPO option may cause unintended consequences if
        any keypoints exist along the cut plane. See Solid Modeling in the
        Modeling and Meshing Guide for an illustration.  See the BOPTN command
        for an explanation of the options available to Boolean operations.
        Element attributes and solid model boundary conditions assigned to the
        original entities will not be transferred to the new entities
        generated.

        Issuing the ASBW command under certain conditions may generate a
        topological degeneracy error. Do not issue the command if:

        A sphere or cylinder has been scaled. (A cylinder must be scaled
        unevenly in the XY plane.)

        A sphere or cylinder has not been scaled but the work plane has been
        rotated.
        """
        command = f"ASBW,{na},{sepo},{keep}"
        return self.run(command, **kwargs)

    def boptn(self, lab="", value="", **kwargs):
        """Specifies Boolean operation options.

        APDL Command: BOPTN

        Parameters
        ----------
        lab
            Default/status key:

            DEFA  - Resets settings to default values.

            STAT  - Lists status of present settings.

        value
            Option settings if Lab = KEEP:

            NO  - Delete entities used as input with a Boolean operation (default).  Entities
                  will not be deleted if meshed or if attached to a higher
                  entity.

            YES  - Keep input solid modeling entities.

        Notes
        -----
        Boolean operations at Revision 5.2 may produce a different number of
        entities than previous revisions of ANSYS.   When running input files
        created at earlier revisions of ANSYS, match the Boolean compatibility
        option (VERSION) to the revision originally used. For instance,  if you
        are running Revision 5.2 and are reading an input file (/INPUT) created
        at Revision 5.1, it is recommended that you set VERSION to RV51 before
        reading the input.

        See the Modeling and Meshing Guide for further details on the functions
        of the RV51 and RV52 labels.

        This command is valid in any processor.
        """
        command = f"BOPTN,{lab},{value}"
        return self.run(command, **kwargs)

    def btol(self, ptol="", **kwargs):
        """Specifies the Boolean operation tolerances.

        APDL Command: BTOL

        Parameters
        ----------
        ptol
            Point coincidence tolerance.  Points within this distance to each
            other will be assumed to be coincident during Boolean operations.
            Loosening the tolerance will increase the run time and storage
            requirements, but will allow more Boolean intersections to succeed.
            Defaults to 0.10E-4.

        Notes
        -----
        Use BTOL,DEFA to reset the setting to its default value.  Use BTOL,STAT
        to list the status of the present setting.
        """
        command = f"BTOL,{ptol}"
        return self.run(command, **kwargs)

    def lcsl(
        self,
        nl1="",
        nl2="",
        nl3="",
        nl4="",
        nl5="",
        nl6="",
        nl7="",
        nl8="",
        nl9="",
        **kwargs,
    ):
        """Divides intersecting lines at their point(s) of intersection.

        APDL Command: LCSL

        Parameters
        ----------
        nl1, nl2, nl3, . . . , nl9
            Numbers of lines to be intersected.  If NL1 = ALL, NL2 to NL9 are
            ignored and the intersection of all selected lines is found.  If
            NL1 = P, use graphical picking to specify lines (NL2 to NL9 are
            ignored).

        Notes
        -----
        Divides intersecting (classifies) lines at their point(s) of
        intersection.  The original lines (and their corresponding keypoint(s))
        will be deleted by default.  See the BOPTN command for the options
        available to Boolean operations.  Element attributes and solid model
        boundary conditions assigned to the original entities will not be
        transferred to the new entities generated.
        """
        command = f"LCSL,{nl1},{nl2},{nl3},{nl4},{nl5},{nl6},{nl7},{nl8},{nl9}"
        return self.run(command, **kwargs)

    def lglue(
        self,
        nl1="",
        nl2="",
        nl3="",
        nl4="",
        nl5="",
        nl6="",
        nl7="",
        nl8="",
        nl9="",
        **kwargs,
    ):
        """Generates new lines by "gluing" lines.

        APDL Command: LGLUE

        Parameters
        ----------
        nl1, nl2, nl3, . . . , nl9
            Numbers of the lines to be glued.  If NL1 = ALL, all selected lines
            will be glued (NL2 to NL9 will be ignored).  If NL1 = P, graphical
            picking is enabled and all remaining command fields are ignored
            (valid only in the GUI).  A component name may also be substituted
            for NL1.

        Notes
        -----
        Use of the LGLUE command generates new lines by "gluing" input lines.
        The glue operation redefines the input lines so that they share
        keypoints at their common ends.  The new lines encompass the same
        geometry as the original lines.  This operation is only valid if the
        intersections of the input lines are keypoints at the ends of those
        lines.  See the Modeling and Meshing Guide for an illustration.  See
        the BOPTN command for an explanation of the options available to
        Boolean operations.  Element attributes and solid model boundary
        conditions assigned to the original entities will not be transferred to
        the new entities generated.

        The LGLUE command results in the merging of keypoints at the common end
        of the lines. The keypoints of the lower numbered line will be kept.
        This means one must be aware of line numbering when multiple LGLUE
        commands are applied to avoid any "ungluing" of geometry.
        """
        command = f"LGLUE,{nl1},{nl2},{nl3},{nl4},{nl5},{nl6},{nl7},{nl8},{nl9}"
        return self.run(command, **kwargs)

    def lina(self, nl="", na="", **kwargs):
        """Finds the intersection of a line with an area.

        APDL Command: LINA

        Parameters
        ----------
        nl
            Number of line to be intersected.  If NL = P, graphical picking is
            enabled and all remaining command fields are ignored (valid only in
            the GUI).

        na
            Number of area to be intersected.

        Notes
        -----
        Finds the intersection of a line with an area.  New lines will be
        generated where the lines intersect the areas.  If the regions of
        intersection are only points, new keypoints will be generated instead.
        See the Modeling and Meshing Guide for an illustration.  See the BOPTN
        command for the options available to Boolean operations.  Element
        attributes and solid model boundary conditions assigned to the original
        entities will not be transferred to the new entities generated.
        """
        command = f"LINA,{nl},{na}"
        return self.run(command, **kwargs)

    def linl(
        self,
        nl1="",
        nl2="",
        nl3="",
        nl4="",
        nl5="",
        nl6="",
        nl7="",
        nl8="",
        nl9="",
        **kwargs,
    ):
        """Finds the common intersection of lines.

        APDL Command: LINL

        Parameters
        ----------
        nl1, nl2, nl3, . . . , nl9
            Numbers of lines to be intersected.  If NL1 = ALL, find the
            intersection of all selected lines and NL2 to NL9 are ignored.  If
            NL1 = P, graphical picking is enabled and all remaining command
            fields are ignored (valid only in the GUI).  A component name may
            also be substituted for NL1.

        Notes
        -----
        Finds the common (not pairwise) intersection of lines.  The common
        intersection is defined as the regions shared (in common) by all lines
        listed on this command.  New lines will be generated where the original
        lines intersect.  If the regions of intersection are only points, new
        keypoints will be generated instead.  See the Modeling and Meshing
        Guide for an illustration.  See the BOPTN command for the options
        available to Boolean operations.  Element attributes and solid model
        boundary conditions assigned to the original entities will not be
        transferred to the new entities generated.
        """
        command = f"LINL,{nl1},{nl2},{nl3},{nl4},{nl5},{nl6},{nl7},{nl8},{nl9}"
        return self.run(command, **kwargs)

    def linp(
        self,
        nl1="",
        nl2="",
        nl3="",
        nl4="",
        nl5="",
        nl6="",
        nl7="",
        nl8="",
        nl9="",
        **kwargs,
    ):
        """Finds the pairwise intersection of lines.

        APDL Command: LINP

        Parameters
        ----------
        nl1, nl2, nl3, . . . , nl9
            Numbers of lines to be intersected pairwise.  If NL1 =  ALL, find
            the pairwise intersection of all selected lines and NL2 to NL9 are
            ignored.  If NL1 = P, graphical picking is enabled and all
            remaining command fields are ignored (valid only in the GUI).  A
            component name may be substituted for NL1.

        Notes
        -----
        Finds the pairwise intersection of lines.  The pairwise intersection is
        defined as any and all regions shared by at least two lines listed on
        this command.  New lines will be generated where the original lines
        intersect pairwise.  If the regions of pairwise intersection are only
        points, new keypoints will be generated.  See the Modeling and Meshing
        Guide for an illustration.  See the BOPTN command for the options
        available to Boolean operations.  Element attributes and solid model
        boundary conditions assigned to the original entities will not be
        transferred to the new entities generated.
        """
        command = f"LINP,{nl1},{nl2},{nl3},{nl4},{nl5},{nl6},{nl7},{nl8},{nl9}"
        return self.run(command, **kwargs)

    def linv(self, nl="", nv="", **kwargs):
        """Finds the intersection of a line with a volume.

        APDL Command: LINV

        Parameters
        ----------
        nl
            Number of line to be intersected.  If NL = P, graphical picking is
            enabled and all remaining command fields are ignored (valid only in
            the GUI).

        nv
            Number of volume to be intersected.

        Notes
        -----
        Finds the intersection of a line with a volume.  New lines will be
        generated where the lines intersect the volumes.  If the regions of
        intersection are only points, new keypoints will be generated instead.
        See the Modeling and Meshing Guide for an illustration.  See the BOPTN
        command for the options available to Boolean operations.  Element
        attributes and solid model boundary conditions assigned to the original
        entities will not be transferred to the new entities generated.
        """
        command = f"LINV,{nl},{nv}"
        return self.run(command, **kwargs)

    def lovlap(
        self,
        nl1="",
        nl2="",
        nl3="",
        nl4="",
        nl5="",
        nl6="",
        nl7="",
        nl8="",
        nl9="",
        **kwargs,
    ):
        """Overlaps lines.

        APDL Command: LOVLAP

        Parameters
        ----------
        nl1, nl2, nl3, . . . , nl9
            Numbers of lines to be overlapped.  If NL1 = ALL, NL2 to NL9 are
            ignored and all selected lines are overlapped.  If NL1 = P,
            graphical picking is enabled and all remaining command fields are
            ignored (valid only in the GUI).  A component name may also be
            substituted for NL1.

        Notes
        -----
        Overlaps lines. Generates new lines which encompass the geometry of all
        the input lines.  The new lines are defined by the regions of
        intersection of the input lines, and by the complementary (non-
        intersecting) regions.  See the Modeling and Meshing Guide for an
        illustration.  This operation is only valid when the region of
        intersection is a line.  See the BOPTN command for an explanation of
        the options available to Boolean operations.  Element attributes and
        solid  model boundary conditions assigned to the original entities will
        not be transferred to the new entities generated.
        """
        command = f"LOVLAP,{nl1},{nl2},{nl3},{nl4},{nl5},{nl6},{nl7},{nl8},{nl9}"
        return self.run(command, **kwargs)

    def lptn(
        self,
        nl1="",
        nl2="",
        nl3="",
        nl4="",
        nl5="",
        nl6="",
        nl7="",
        nl8="",
        nl9="",
        **kwargs,
    ):
        """Partitions lines.

        APDL Command: LPTN

        Parameters
        ----------
        nl1, nl2, nl3, . . . , nl9
            Numbers of lines to be operated on.  If NL1 = ALL, NL2 to NL9 are
            ignored all selected lines are used.  If NL1 = P, graphical picking
            is enabled and all remaining command fields are ignored (valid only
            in the GUI).  A component name may be substituted for NL1.

        Notes
        -----
        Partitions lines.  Generates new lines which encompass the geometry of
        all the input lines.  The new lines are defined by both the regions of
        intersection of the input lines and the complementary (non-
        intersecting) regions.  See the Modeling and Meshing Guide for an
        illustration. See the BOPTN command for an explanation of the options
        available to Boolean operations.  Element attributes and solid model
        boundary conditions assigned to the original entities will not be
        transferred to the new entities generated.
        """
        command = f"LPTN,{nl1},{nl2},{nl3},{nl4},{nl5},{nl6},{nl7},{nl8},{nl9}"
        return self.run(command, **kwargs)

    def lsba(self, nl="", na="", sepo="", keepl="", keepa="", **kwargs):
        """Subtracts areas from lines.

        APDL Command: LSBA

        Parameters
        ----------
        nl
            Line (or lines, if picking is used) to be subtracted from.  If ALL,
            use all selected lines.  If NL = P, graphical picking is enabled
            and all remaining command fields are ignored (valid only in the
            GUI).  A component name may also be substituted for NL.

        na
            Area (or areas, if picking is used) to be subtracted.  If ALL, use
            all selected areas.  A component name may also be substituted for
            NA.

        sepo
            Behavior if the intersection of the lines and the areas is a
            keypoint or keypoints:

            (blank) - The resulting lines will share keypoint(s) where they touch.

            SEPO - The resulting lines will have separate, but coincident keypoint(s) where they
                   touch.

        keepl
            Specifies whether NL lines are to be deleted:

            (blank) - Use the setting of KEEP on the BOPTN command.

            DELETE - Delete NL lines after LSBA operation (override BOPTN command settings).

            KEEP - Keep NL lines after LSBA operation (override BOPTN command settings).

        keepa
            Specifies whether NA areas are to be deleted:

            (blank) - Use the setting of KEEP on the BOPTN command.

            DELETE - Delete areas after LSBA operation (override BOPTN command settings).

            KEEP - Keep areas after LSBA operation (override BOPTN command settings).

        Notes
        -----
        Generates new lines by subtracting the regions common to both NL lines
        and NA areas (the intersection) from the NL lines.  The intersection
        can be a line(s) or keypoint(s).  If the intersection is a keypoint and
        SEPO is blank, the NL line is divided at the keypoint and the resulting
        lines will be connected, sharing a common keypoint where they touch.
        If SEPO is set to SEPO, NL is divided into two unconnected lines with
        separate keypoints where they touch.  See the Modeling and Meshing
        Guide for an illustration.  See the BOPTN command for an explanation of
        the options available to Boolean operations.  Element attributes and
        solid model boundary conditions assigned to the original entities will
        not be transferred to the new entities generated.
        """
        command = f"LSBA,{nl},{na},{sepo},{keepl},{keepa}"
        return self.run(command, **kwargs)

    def lsbl(self, nl1="", nl2="", sepo="", keep1="", keep2="", **kwargs):
        """Subtracts lines from lines.

        APDL Command: LSBL

        Parameters
        ----------
        nl1
            Line (or lines, if picking is used) to be subtracted from.  If ALL,
            use all selected lines.  Lines specified in this argument are not
            available for use in the NL2 argument.  If P, graphical picking is
            enabled (valid only in the GUI) and all remaining fields are
            ignored.  A component name may also be substituted for NL1.

        nl2
            Line (or lines, if picking is used) to subtract.  If ALL, use all
            selected lines (except those included in the NL1 argument).  A
            component name may also be substituted for NL2.

        sepo
            Behavior if the intersection of the NL1 lines and the NL2 lines is
            a keypoint or keypoints:

            (blank) - The resulting lines will share keypoint(s) where they touch.

            SEPO - The resulting lines will have separate, but coincident keypoint(s) where they
                   touch.

        keep1
            Specifies whether NL1 lines are to be deleted:

            (blank) - Use the setting of KEEP on the BOPTN command.

            DELETE - Delete NL1 lines after LSBL operation (override BOPTN command settings).

            KEEP - Keep NL1 lines after LSBL operation (override BOPTN command settings).

        keep2
            Specifies whether NL2 lines are to be deleted:

            (blank) - Use the setting of KEEP on the BOPTN command.

            DELETE - Delete NL2 lines after LSBL operation (override BOPTN command settings).

            KEEP - Keep NL2 lines after LSBL operation (override BOPTN command settings).

        Notes
        -----
        Generates new lines by subtracting the regions common to both NL1 and
        NL2 lines (the intersection) from the NL1 lines.  The intersection can
        be a line(s) or point(s).  If the intersection is a point and SEPO is
        blank, the NL1 line is divided at the point and the resulting lines
        will be connected, sharing a common keypoint where they touch.  If SEPO
        is set to SEPO, NL1 is divided into two unconnected lines with separate
        keypoints where they touch.  See the Modeling and Meshing Guide for an
        illustration.  See the BOPTN command for an explanation of the options
        available to Boolean operations.  Element attributes and solid model
        boundary conditions assigned to the original entities will not be
        transferred to the new entities generated.  LSBL,ALL,ALL will have no
        effect since all the lines (in NL1) will be unavailable as NL2 lines.
        """
        command = f"LSBL,{nl1},{nl2},{sepo},{keep1},{keep2}"
        return self.run(command, **kwargs)

    def lsbv(self, nl="", nv="", sepo="", keepl="", keepv="", **kwargs):
        """Subtracts volumes from lines.

        APDL Command: LSBV

        Parameters
        ----------
        nl
            Line (or lines, if picking is used) to be subtracted from.  If ALL,
            use all selected lines.  If NL = P, graphical picking is enabled
            and all remaining command fields are ignored (valid only in the
            GUI).  A component name may also be substituted for NL.

        nv
            Volume (or volumes, if picking is used) to be subtracted.  If ALL,
            use all selected volumes.  A component name may also be substituted
            for NV.

        sepo
            Behavior if the intersection of the NL lines and the NV volumes is
            a keypoint or keypoints:

            (blank) - The resulting lines will share keypoint(s) where they touch.

            SEPO - The resulting lines will have separate, but coincident keypoint(s) where they
                   touch.

        keepl
            Specifies whether NL lines are to be deleted:

            (blank) - Use the setting of KEEP on the BOPTN command.

            DELETE - Delete NL lines after LSBV operation (override BOPTN command settings).

            KEEP - Keep NL lines after LSBV operation (override BOPTN command settings).

        keepv
            Specifies whether NV volumes are to be deleted:

            (blank) - Use the setting of KEEP on the BOPTN command.

            DELETE - Delete NV volumes after LSBV operation (override BOPTN command settings).

            KEEP - Keep NV volumes after LSBV operation (override BOPTN command settings).

        Notes
        -----
        Generates new lines by subtracting the regions common to both NL lines
        and NV volumes (the intersection) from the NL lines.  The intersection
        can be a line(s) or point(s).  If the intersection is a point and SEPO
        is blank, the NL1 line is divided at the point and the resulting lines
        will be connected, sharing a common keypoint where they touch.  If SEPO
        is set to SEPO, NL1 is divided into two unconnected lines with separate
        keypoints where they touch.  See the Modeling and Meshing Guide for an
        illustration.  See the BOPTN command for an explanation of the options
        available to Boolean operations.  Element attributes and solid model
        boundary conditions assigned to the original entities will not be
        transferred to the new entities generated.  LSBL,ALL,ALL will have no
        effect since all the lines (in NL1) will be unavailable as NL2 lines.
        """
        command = f"LSBV,{nl},{nv},{sepo},{keepl},{keepv}"
        return self.run(command, **kwargs)

    def lsbw(self, nl="", sepo="", keep="", **kwargs):
        """Subtracts the intersection of the working plane from lines (divides

        APDL Command: LSBW
        lines).

        Parameters
        ----------
        nl
            Line (or lines, if picking is used) to be subtracted from.  If NL =
            ALL, use all selected lines.  If NL = P, graphical picking is
            enabled and all remaining command fields are ignored (valid only in
            the GUI).  A component name may also be input for NL.

        sepo
            Behavior of the created boundary.

            (blank) - The resulting lines will share keypoint(s) where they touch.

            SEPO - The resulting lines will have separate, but coincident keypoint(s).

        keep
            Specifies whether NL lines are to be deleted:

            (blank) - Use the setting of KEEP on the BOPTN command.

            DELETE - Delete NL lines after LSBW operation (override BOPTN command settings).

            KEEP - Keep NL lines after LSBW operation (override BOPTN command settings).

        Notes
        -----
        Generates new lines by subtracting the intersection of the working
        plane from the NL lines.  The intersection will be a keypoint(s).  The
        working plane must not be in the same plane as the NL line(s).  If SEPO
        is blank, the NL line is divided and the resulting lines will be
        connected, sharing a common keypoint where they touch.  If SEPO is set
        to SEPO, NL is divided into two unconnected lines with separate
        keypoints.  See the Modeling and Meshing Guide for an illustration.
        See the BOPTN command for an explanation of the options available to
        Boolean operations.  Element attributes and solid model boundary
        conditions assigned to the original entities will not be transferred to
        the new entities generated.  Areas that completely contain the input
        lines will be updated if the lines are divided by this operation.
        """
        command = f"LSBW,{nl},{sepo},{keep}"
        return self.run(command, **kwargs)

    def vadd(
        self,
        nv1="",
        nv2="",
        nv3="",
        nv4="",
        nv5="",
        nv6="",
        nv7="",
        nv8="",
        nv9="",
        **kwargs,
    ):
        """Adds separate volumes to create a single volume.

        APDL Command: VADD

        Parameters
        ----------
        nv1, nv2, nv3, . . . , nv9
            Numbers of volumes to be added.  If NV1 = ALL, add all selected
            volumes and ignore NV2 to NV9.  If NV1 = P, graphical picking is
            enabled and all remaining command fields are ignored (valid only in
            the GUI).  A component name may also be substituted for NV1.

        Notes
        -----
        Adds separate volumes to create a single volume.  The original volumes
        (and their corresponding areas, lines and keypoints) will be deleted by
        default [BOPTN].  See the BOPTN command for the options available to
        Boolean operations.  Element attributes and solid model boundary
        conditions assigned to the original entities will not be transferred to
        the new entities generated.  Concatenated entities are not valid with
        this command.
        """
        command = f"VADD,{nv1},{nv2},{nv3},{nv4},{nv5},{nv6},{nv7},{nv8},{nv9}"
        return self.run(command, **kwargs)

    def vglue(
        self,
        nv1="",
        nv2="",
        nv3="",
        nv4="",
        nv5="",
        nv6="",
        nv7="",
        nv8="",
        nv9="",
        **kwargs,
    ):
        """Generates new volumes by "gluing" volumes.

        APDL Command: VGLUE

        Parameters
        ----------
        nv1, nv2, nv3, . . . , nv9
            Numbers of the volumes to be glued.  If NV1 = ALL, all selected
            volumes will be glued (NV2 to NV9 will be ignored).  If NV1 = P,
            graphical picking is enabled and all remaining command fields are
            ignored (valid only in the GUI).  A component name may also be
            substituted for NV1.

        Notes
        -----
        Use of the VGLUE command generates new volumes by "gluing" input
        volumes.  The glue operation redefines the input volumes so that they
        share areas along their common boundaries.  The new volumes encompass
        the same geometry as the original volumes.  This operation is only
        valid if the intersections of the input volumes are areas along the
        boundaries of those volumes.  See the Modeling and Meshing Guide for an
        illustration.  See the BOPTN command for an explanation of the options
        available to Boolean operations.  Element attributes and solid model
        boundary conditions assigned to the original entities will not be
        transferred to the new entities generated.

        The VGLUE command results in the merging of areas, lines, and keypoints
        at the common volume boundaries. The areas, lines, and keypoints of the
        lower numbered volume will be kept. This means one must be aware of
        volume numbering when multiple VGLUE commands are applied to avoid any
        "ungluing" of geometry.
        """
        command = f"VGLUE,{nv1},{nv2},{nv3},{nv4},{nv5},{nv6},{nv7},{nv8},{nv9}"
        return self.run(command, **kwargs)

    def vinp(
        self,
        nv1="",
        nv2="",
        nv3="",
        nv4="",
        nv5="",
        nv6="",
        nv7="",
        nv8="",
        nv9="",
        **kwargs,
    ):
        """Finds the pairwise intersection of volumes.

        APDL Command: VINP

        Parameters
        ----------
        nv1, nv2, nv3, . . . , nv9
            Numbers of volumes to be intersected pairwise.  If NV1 = ALL, NV2
            to NV9 are ignored and the pairwise intersection of all selected
            volumes is found.  If NV1 = P, graphical picking is enabled and all
            remaining command fields are ignored (valid only in the GUI).  A
            component name may also be substituted for NV1.

        Notes
        -----
        Finds the pairwise intersection of volumes.  The pairwise intersection
        is defined as all regions shared by any two or more volumes listed on
        this command.  New volumes will be generated where the original volumes
        intersect pairwise.  If the regions of pairwise intersection are only
        areas, new areas will be generated.  See the Modeling and Meshing Guide
        for an illustration.  See the BOPTN command for an explanation of the
        options available to Boolean operations.  Element attributes and solid
        model boundary conditions assigned to the original entities will not be
        transferred to the new entities generated.
        """
        command = f"VINP,{nv1},{nv2},{nv3},{nv4},{nv5},{nv6},{nv7},{nv8},{nv9}"
        return self.run(command, **kwargs)

    def vinv(
        self,
        nv1="",
        nv2="",
        nv3="",
        nv4="",
        nv5="",
        nv6="",
        nv7="",
        nv8="",
        nv9="",
        **kwargs,
    ):
        """Finds the intersection of volumes.

        APDL Command: VINV

        Parameters
        ----------
        nv1, nv2, nv3, . . . , nv9
            Numbers of volumes to be intersected.  If NV1 = ALL, NV2 to NV9 are
            ignored, and the intersection of all selected volumes is found.  If
            NV1 = P, graphical picking is enabled and all remaining command
            fields are ignored (valid only in the GUI).  A component name may
            also be substituted for NV1.

        Notes
        -----
        Finds the common (not pairwise) intersection of volumes.  The common
        intersection is defined as the regions shared (in common) by all
        volumes listed on this command.  New volumes will be generated where
        the original volumes intersect.  If the regions of intersection are
        only areas, new areas will be generated instead.  See the Modeling and
        Meshing Guide for an illustration.  See the BOPTN command for an
        explanation of the options available to Boolean operations.  Element
        attributes and solid model boundary conditions assigned to the original
        entities will not be transferred to the new entities generated.
        """
        command = f"VINV,{nv1},{nv2},{nv3},{nv4},{nv5},{nv6},{nv7},{nv8},{nv9}"
        return self.run(command, **kwargs)

    def vovlap(
        self,
        nv1="",
        nv2="",
        nv3="",
        nv4="",
        nv5="",
        nv6="",
        nv7="",
        nv8="",
        nv9="",
        **kwargs,
    ):
        """Overlaps volumes.

        APDL Command: VOVLAP

        Parameters
        ----------
        nv1, nv2, nv3, . . . , nv9
            Numbers of volumes to be operated on.  If NV1 = ALL, NV2 to NV9 are
            ignored and all selected volumes are used.  If NV1 = P, graphical
            picking is enabled and all remaining command fields are ignored
            (valid only in the GUI).  A component name may also be substituted
            for NV1.

        Notes
        -----
        Overlaps volumes. Generates new volumes which encompass the geometry of
        all  the input volumes.  The new volumes are defined by the regions of
        intersection  of the input volumes, and by the complementary (non-
        intersecting) regions.   See the Modeling and Meshing Guide for an
        illustration.  This operation is only valid  when the region of
        intersection is a volume.  See the BOPTN command for an explanation of
        the options available to Boolean operations.  Element attributes and
        solid model boundary conditions assigned to the original entities will
        not be transferred to the new entities generated.
        """
        command = f"VOVLAP,{nv1},{nv2},{nv3},{nv4},{nv5},{nv6},{nv7},{nv8},{nv9}"
        return self.run(command, **kwargs)

    def vptn(
        self,
        nv1="",
        nv2="",
        nv3="",
        nv4="",
        nv5="",
        nv6="",
        nv7="",
        nv8="",
        nv9="",
        **kwargs,
    ):
        """Partitions volumes.

        APDL Command: VPTN

        Parameters
        ----------
        nv1, nv2, nv3, . . . , nv9
            Numbers of volumes to be operated on.  If NV1 = ALL, NV2 to NV9 are
            ignored and all selected volumes are used.  If NV1  = P, graphical
            picking is enabled and all remaining command fields are ignored
            (valid only in the GUI).  A component name may also be substituted
            for NV1.

        Notes
        -----
        Partitions volumes.  Generates new volumes which encompass the geometry
        of all the input volumes.  The new volumes are defined by the regions
        of intersection of the input volumes, and by the complementary (non-
        intersecting) regions.  See the Modeling and Meshing Guide for an
        illustration. See the BOPTN command for an explanation of the options
        available to Boolean operations.  Element attributes and solid model
        boundary conditions assigned to the original entities will not be
        transferred to the new entities generated.
        """
        command = f"VPTN,{nv1},{nv2},{nv3},{nv4},{nv5},{nv6},{nv7},{nv8},{nv9}"
        return self.run(command, **kwargs)

    def vsba(self, nv="", na="", sepo="", keepv="", keepa="", **kwargs):
        """Subtracts areas from volumes.

        APDL Command: VSBA

        Parameters
        ----------
        nv
            Volume (or volumes, if picking is used) to be subtracted from.  If
            ALL, use all selected volumes.  If P, graphical picking is enabled
            (valid only in the GUI) and remaining fields are ignored.  A
            component name may also be substituted for NV.

        na
            Area (or areas, if picking is used) to subtract.  If ALL, use all
            selected areas.  A component name may also be substituted for NA.

        sepo
            Behavior of the touching boundary:

            (blank) - The resulting volumes will share area(s) where they touch.

            SEPO - The resulting volumes will have separate, but coincident area(s) where they
                   touch.

        keepv
            Specifies whether NV volumes are to be deleted:

            (blank) - Use the setting of KEEP on the BOPTN command.

            DELETE - Delete NV volumes after VSBA operation (override BOPTN command settings).

            KEEP - Keep NV volumes after VSBA operation (override BOPTN command settings).

        keepa
            Specifies whether NA areas are to be deleted:

            (blank) - Use the setting of KEEP on the BOPTN command.

            DELETE - Delete NA areas after VSBA operation (override BOPTN command settings).

            KEEP - Keep NA areas after VSBA operation (override BOPTN command settings).

        Notes
        -----
        Generates new volumes by subtracting the regions common to both the
        volumes and areas (the intersection) from the NV volumes.  The
        intersection will be an area(s).  If SEPO is blank, the volume is
        divided at the area and the resulting volumes will be connected,
        sharing a common area where they touch.  If SEPO is set to SEPO, the
        volume is divided into two unconnected volumes with separate areas
        where they touch.  See the Modeling and Meshing Guide for an
        illustration.  See the BOPTN command for an explanation of the options
        available to Boolean operations.  Element attributes and solid model
        boundary conditions assigned to the original entities will not be
        transferred to the new entities generated.
        """
        command = f"VSBA,{nv},{na},{sepo},{keepv},{keepa}"
        return self.run(command, **kwargs)

    def vsbv(self, nv1="", nv2="", sepo="", keep1="", keep2="", **kwargs):
        """Subtracts volumes from volumes.

        APDL Command: VSBV

        Parameters
        ----------
        nv1
            Volume (or volumes, if picking is used) to be subtracted
            from.  If ALL, use all selected volumes.  Volumes
            specified in set NV2 are removed from set NV1.  If P,
            graphical picking is enabled (valid only in the GUI) and
            remaining fields are ignored.  A component name may also
            be substituted for NV1.

        nv2
            Volume (or volumes, if picking is used) to subtract.  If
            ALL, use all selected volumes (except those included in
            the NV1 argument).  A component name may also be
            substituted for NV2.

        sepo
            Behavior if the intersection of the NV1 volumes and the
            NV2 volumes is an area or areas:

            (blank) - The resulting volumes will share area(s) where they touch.

            SEPO - The resulting volumes will have separate, but
                   coincident area(s) where they touch.

        keep1
            Specifies whether NV1 volumes are to be deleted:

            (blank) - Use the setting of KEEP on the BOPTN command.

            DELETE - Delete NV1 volumes after VSBV operation (override
                     BOPTN command settings).

            KEEP - Keep NV1 volumes after VSBV operation (override
                   BOPTN command settings).

        keep2
            Specifies whether NV2 volumes are to be deleted:

            (blank) - Use the setting of KEEP on the BOPTN command.

            DELETE - Delete NV2 volumes after VSBV operation (override
                     BOPTN command settings).

            KEEP - Keep NV2 volumes after VSBV operation (override
                   BOPTN command settings).

        Notes
        -----
        Generates new volumes by subtracting the regions common to
        both NV1 and NV2 volumes (the intersection) from the NV1
        volumes.  The intersection can be a volume(s) or area(s).  If
        the intersection is an area and SEPO is blank, the NV1 volume
        is divided at the area and the resulting volumes will be
        connected, sharing a common area where they touch.  If SEPO is
        set to SEPO, NV1 is divided into two unconnected volumes with
        separate areas where they touch.  See the Modeling and Meshing
        Guide for an illustration.  See the BOPTN command for an
        explanation of the options available to Boolean operations.
        Element attributes and solid model boundary conditions
        assigned to the original entities will not be transferred to
        the new entities generated.  VSBV,ALL,ALL will have no effect
        because all the volumes in set NV1will have been moved to set
        NV2.
        """
        command = f"VSBV,{nv1},{nv2},{sepo},{keep1},{keep2}"
        return self.run(command, **kwargs)

    def vsbw(self, nv="", sepo="", keep="", **kwargs):
        """Subtracts intersection of the working plane from volumes.

        APDL Command: VSBW

        Parameters
        ----------
        nv
            Volume (or volumes, if picking is used) to be subtracted from.  If
            NV = ALL, use all selected volumes.  If NV = P, graphical picking
            is enabled (valid only in the GUI).  A component name may also be
            input for NV.

        sepo
            Behavior of the created boundary.

            (blank) - The resulting volumes will share area(s) where they touch.

            SEPO - The resulting volumes will have separate, but coincident area(s).

        keep
            Specifies whether NV volumes are to be deleted.

            (blank) - Use the setting of KEEP on the BOPTN command.

            DELETE - Delete NV volumes after VSBW operation (override BOPTN command settings).

            KEEP - Keep NV volumes after VSBW operation (override BOPTN command settings).

        Notes
        -----
        Generates new volumes by subtracting the intersection of the working
        plane from the NV volumes.  The intersection will be an area(s).  If
        SEPO is blank, the volume is divided at the area and the resulting
        volumes will be connected, sharing a common area where they touch.  If
        SEPO is set to SEPO, the volume is divided into two unconnected volumes
        with separate areas.  The SEPO option may cause unintended consequences
        if any keypoints exist along the cut plane. See the Modeling and
        Meshing Guide for an illustration.  See the BOPTN command for an
        explanation of the options available to Boolean operations.  Element
        attributes and solid model boundary conditions assigned to the original
        entities will not be transferred to the new entities generated.

        Issuing the VSBW command under certain conditions may generate a
        topological degeneracy error. Do not issue the command if:

        A sphere or cylinder has been scaled. (A cylinder must be scaled
        unevenly in the XY plane.)

        A sphere or cylinder has not been scaled but the work plane has been
        rotated.
        """
        command = f"VSBW,{nv},{sepo},{keep}"
        return self.run(command, **kwargs)



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Store parameters for a PyMAPDL-specific theme for PyVista"""

import numpy as np

try:
    from matplotlib.colors import LinearSegmentedColormap, ListedColormap

    _HAS_MATPLOTLIB = True
except ModuleNotFoundError:
    _HAS_MATPLOTLIB = False


from ansys.mapdl.core import _HAS_PYVISTA

if _HAS_PYVISTA:
    from cycler import Cycler
    from pyvista.plotting.colors import get_cycler

    try:
        from pyvista.plotting.themes import Theme

    except ImportError:
        from pyvista import __version__ as pyvista_version

        if "0.40" in pyvista_version:
            from pyvista.themes import Theme
        else:  # older versions
            from pyvista.themes import DefaultTheme as Theme

else:  # pragma: no cover

    class Theme:
        pass


MAPDL_colorbar = (
    np.array(
        [
            [0, 0, 255],
            [0, 178, 255],
            [0, 255, 255],
            [0, 255, 178],
            [0, 255, 0],
            [178, 255, 0],
            [255, 255, 0],
            [255, 178, 0],
            [255, 0, 0],
        ],
        dtype=float,
    )
    / 255
)

if _HAS_MATPLOTLIB:
    PyMAPDL_cmap: ListedColormap = ListedColormap(MAPDL_colorbar, name="PyMAPDL")


def get_ansys_cmap(N: int = 9) -> "LinearSegmentedColormap":
    """Returns a Matplotlib colormap of given length.

    Returns a Matplotlib colormap which is the result of interpolate the MAPDL
    colormap to obtain a colormap of a given number of colors.

    Parameters
    ----------
    N : int, optional
        Number of colours, by default 9

    Returns
    -------
    matplotlib.colors.LinearSegmentedColormap
        Colormap
    """
    return LinearSegmentedColormap.from_list("PyMAPDL", MAPDL_colorbar.tolist(), N=N)


def get_ansys_colors(N: int = 9) -> np.array:
    """Get N number of colors as array.

    Obtain N unique colors as a Numpy array (N x 4). Transparency is included.
    This function uses `get_ansys_cmap` to interpolate between the colors given
    in `PyMAPDL_cmap`.

    Parameters
    ----------
    N : int, optional
        Number of colors, by default 9

    Returns
    -------
    np.array
        Color array

    Raises
    ------
    ModuleNotFoundError
        Matplotlib can not be found.
    """
    if not _HAS_MATPLOTLIB:
        raise ModuleNotFoundError(
            "'matplotlib' package is needed for 'get_ansys_colors'."
        )

    cmap = get_ansys_cmap(N=N)
    return cmap([i for i in range(N)])


def get_ansys_color_cycle(N: int = 9) -> np.array:
    """Get a color cycler

    Give an array of N colors which is the result of cycling through the MAPDL
    colormap. Only m unique colors can be provided through this function, where
    m is the minimal of 9 or N.

    Parameters
    ----------
    N : int, optional
        Number of colors, by default 9

    Returns
    -------
    np.array
        Array of colors
    """
    from cycler import cycler

    cyc_ = cycler(color=MAPDL_colorbar.tolist())()
    return np.array([each["color"] for i, each in zip(range(N), cyc_)])


class MapdlTheme(Theme):
    """Provides the PyMAPDL-specific theme for PyVista.

    The theme includes these defaults:

    - PyMAPDL colormap as default colormap.
    - ``'Courier'`` font family for an interactive plot
    - ``'PyMAPDL'`` for the title of an interactive plot

    Examples
    --------
    Create a custom theme with unique parameters from the ``MapdlTheme`` base.

    >>> from ansys.mapdl import core as pymapdl
    >>> my_theme = pymapdl.MapdlTheme()
    >>> my_theme.background = 'white'
    >>> my_theme.axes.show = False
    >>> my_theme.show_scalar_bar = False

    Apply this theme to element plotting.

    >>> mapdl.eplot(theme=theme)

    Apply this theme to area plotting.

    >>> mapdl.aplot(theme=theme)

    """

    def __init__(self):
        """Initialize the theme."""
        super().__init__()

        self.name: str = "PyMAPDL-Theme"
        self.title: str = "PyMAPDL"
        self.background: str = "paraview"
        self.interactive: bool = True

        if _HAS_MATPLOTLIB:
            self.cmap = PyMAPDL_cmap
        self.font.family: str = "arial"

        self.font.size: int = 18
        self.font.title_size: int = 18
        self.font.label_size: int = 18
        self.font.color: str = "black"

        self.axes.x_color: str = "tomato"
        self.axes.y_color: str = "seagreen"
        self.axes.z_color: str = "blue"

        self.show_edges: bool = False
        self.color: str = "lightblue"
        self.outline_color: str = "black"
        self.edge_color: str = "black"

        self.color_cycler: Cycler = get_cycler(MAPDL_colorbar.tolist())
        self.render_points_as_spheres: bool = True


def _apply_default_theme():
    import pyvista as pv

    pv.global_theme = MapdlTheme()
